@@ -4,16 +4,17 @@ from torch import optim
 from torch import nn
 import torch.utils.data
 from torch.nn import BatchNorm1d, Dropout, LeakyReLU, Linear, Module, ReLU, Sequential, Sigmoid
+import warnings
 
 import opacus
 
-from ctgan.data_sampler import DataSampler
+from .data_sampler import DataSampler
 from ctgan.data_transformer import DataTransformer
 from ctgan.synthesizers import CTGANSynthesizer
 
 
 class Discriminator(Module):
-    def __init__(self, input_dim, discriminator_dim, loss, pac=10):
+    def __init__(self, input_dim, discriminator_dim, loss, pac=1):
         super(Discriminator, self).__init__()
         torch.cuda.manual_seed(0)
         torch.manual_seed(0)
@@ -33,7 +34,7 @@ class Discriminator(Module):
             seq += [Sigmoid()]
         self.seq = Sequential(*seq)
 
-    def calc_gradient_penalty(self, real_data, fake_data, device='cpu', pac=10, lambda_=10):
+    def calc_gradient_penalty(self, real_data, fake_data, device='cpu', pac=1, lambda_=10):
         alpha = torch.rand(real_data.size(0) // pac, 1, 1, device=device)
         alpha = alpha.repeat(1, pac, real_data.size(1))
         alpha = alpha.view(-1, real_data.size(1))
@@ -100,8 +101,6 @@ def _custom_create_or_extend_grad_sample(
     This custom code will not work when using optimizer.virtual_step()
     """
 
-    # print ("now this happen")
-
     if hasattr(param, "grad_sample"):
         param.grad_sample = param.grad_sample + grad_sample
         # param.grad_sample = torch.cat((param.grad_sample, grad_sample), batch_dim)
@@ -121,17 +120,18 @@ class DPCTGAN(CTGANSynthesizer):
                  discriminator_decay=1e-6,
                  batch_size=500,
                  discriminator_steps=1,
-                 log_frequency=True,
+                 log_frequency=False,
                  verbose=True,
                  epochs=300,
-                 pac=10,
+                 pac=1,
                  cuda=True,
                  disabled_dp=False,
-                 target_delta=None,
+                 delta=None,
                  sigma=5,
                  max_per_sample_grad_norm=1.0,
                  epsilon=1,
-                 loss="cross_entropy",):
+                 loss="cross_entropy",
+                 category_epsilon_pct=0.1):
 
         assert batch_size % 2 == 0
 
@@ -149,12 +149,13 @@ class DPCTGAN(CTGANSynthesizer):
         self._log_frequency = log_frequency
         self._verbose = verbose
         self._epochs = epochs
+        self._category_epsilon_pct = category_epsilon_pct
         self.pac = pac
 
         # opacus parameters
         self.sigma = sigma
         self.disabled_dp = disabled_dp
-        self.target_delta = target_delta
+        self.delta = delta
         self.max_per_sample_grad_norm = max_per_sample_grad_norm
         self.epsilon = epsilon
         self.epsilon_list = []
@@ -179,23 +180,56 @@ class DPCTGAN(CTGANSynthesizer):
 
         if self.loss != "cross_entropy":
             # Monkeypatches the _create_or_extend_grad_sample function when calling opacus
-            opacus.supported_layers_grad_samplers._create_or_extend_grad_sample = (
+            opacus.grad_sample.utils.create_or_extend_grad_sample = (
                 _custom_create_or_extend_grad_sample
             )
 
+        if self._log_frequency:
+            warnings.warn(
+                "log_frequency is selected.  This may result in oversampling frequent "
+                "categories, which could cause privacy leaks."
+            )
+
     def train(self, data, categorical_columns=None, ordinal_columns=None, update_epsilon=None):
         if update_epsilon:
             self.epsilon = update_epsilon
 
+        for col in categorical_columns:
+            if str(data[col].dtype).startswith('float'):
+                raise ValueError(
+                    "It looks like you are passing in a vector of continuous values"
+                    f"to a categorical column at [{col}]."
+                    "Please discretize and pass in categorical columns with"
+                    "unsigned integer or string category names."
+                )
+
         self._transformer = DataTransformer()
         self._transformer.fit(data, discrete_columns=categorical_columns)
 
         train_data = self._transformer.transform(data)
 
+        sampler_eps = 0.0
+        if categorical_columns and self._category_epsilon_pct:
+            sampler_eps = self.epsilon * self._category_epsilon_pct
+            per_col_sampler_eps = sampler_eps / len(categorical_columns)
+            self.epsilon = self.epsilon - sampler_eps
+        else:
+            per_col_sampler_eps = None
+
         self._data_sampler = DataSampler(
             train_data,
             self._transformer.output_info_list,
-            self._log_frequency)
+            self._log_frequency,
+            per_column_epsilon=per_col_sampler_eps)
+
+        spent = self._data_sampler.total_spent
+        if (
+            spent > sampler_eps
+            and not np.isclose(spent, sampler_eps)
+        ):
+            raise AssertionError(
+                f"The data sampler used {spent} epsilon and was budgeted for {sampler_eps}"
+            )
 
         data_dim = self._transformer.output_dimensions
 
@@ -251,6 +285,33 @@ class DPCTGAN(CTGANSynthesizer):
 
         steps_per_epoch = max(len(train_data) // self._batch_size, 1)
         for i in range(self._epochs):
+            if not self.disabled_dp:
+                # if self.loss == 'cross_entropy':
+                #    autograd_grad_sample.clear_backprops(discriminator)
+                # else:
+                for p in discriminator.parameters():
+                    if hasattr(p, "grad_sample"):
+                        del p.grad_sample
+
+                if self.delta is None:
+                    self.delta = 1 / (train_data.shape[0] * np.sqrt(train_data.shape[0]))
+
+                epsilon, best_alpha = optimizerD.privacy_engine.get_privacy_spent(
+                    self.delta
+                )
+
+                self.epsilon_list.append(epsilon)
+                self.alpha_list.append(best_alpha)
+                if self.epsilon < epsilon:
+                    if self._epochs == 1:
+                        raise ValueError(
+                                    "Inputted epsilon and sigma parameters are too small to"
+                                    + " create a private dataset. Try increasing either parameter "
+                                    + "and rerunning."
+                                )
+                    else:
+                        break
+
             for id_ in range(steps_per_epoch):
                 fakez = torch.normal(mean=mean, std=std)
 
@@ -380,28 +441,6 @@ class DPCTGAN(CTGANSynthesizer):
                 loss_g.backward()
                 optimizerG.step()
 
-                if not self.disabled_dp:
-                    # if self.loss == 'cross_entropy':
-                    #    autograd_grad_sample.clear_backprops(discriminator)
-                    # else:
-                    for p in discriminator.parameters():
-                        if hasattr(p, "grad_sample"):
-                            del p.grad_sample
-
-                    if self.target_delta is None:
-                        self.target_delta = 1 / train_data.shape[0]
-
-                    epsilon, best_alpha = optimizerD.privacy_engine.get_privacy_spent(
-                        self.target_delta
-                    )
-
-                    self.epsilon_list.append(epsilon)
-                    self.alpha_list.append(best_alpha)
-                    # if self.verbose:
-
-            if not self.disabled_dp:
-                if self.epsilon < epsilon:
-                    break
             self.loss_d_list.append(loss_d)
             self.loss_g_list.append(loss_g)
             if self.verbose:
@@ -428,12 +467,12 @@ class DPCTGAN(CTGANSynthesizer):
             std = mean + 1
             fakez = torch.normal(mean=mean, std=std).to(self._device)
 
-            condvec = self._data_sampler.sample_original_condvec(self._batch_size)
+            condvec = self._data_sampler.sample_condvec(self._batch_size)
 
             if condvec is None:
                 pass
             else:
-                c1 = condvec
+                c1, m1, col, opt = condvec
                 c1 = torch.from_numpy(c1).to(self._device)
                 fakez = torch.cat([fakez, c1], dim=1)
 
