@@ -6,8 +6,9 @@ from torch import nn
 import torch.utils.data
 from torch.nn import BatchNorm1d, Dropout, LeakyReLU, Linear, Module, ReLU, Sequential, Sigmoid
 from torch.autograd import Variable
+import warnings
 
-from ctgan.data_sampler import DataSampler
+from .data_sampler import DataSampler
 from ctgan.data_transformer import DataTransformer
 from ctgan.synthesizers import CTGANSynthesizer
 
@@ -107,7 +108,7 @@ class PATECTGAN(CTGANSynthesizer):
                  discriminator_decay=1e-6,
                  batch_size=500,
                  discriminator_steps=1,
-                 log_frequency=True,
+                 log_frequency=False,
                  verbose=False,
                  epochs=300,
                  pac=1,
@@ -119,9 +120,10 @@ class PATECTGAN(CTGANSynthesizer):
                  teacher_iters=5,
                  student_iters=5,
                  sample_per_teacher=1000,
-                 delta=1e-5,
+                 delta=None,
                  noise_multiplier=1e-3,
-                 moments_order=100
+                 moments_order=100,
+                 category_epsilon_pct=0.1
                  ):
 
         assert batch_size % 2 == 0
@@ -143,6 +145,8 @@ class PATECTGAN(CTGANSynthesizer):
         self.pac = pac
         self.epsilon = epsilon
 
+        self._category_epsilon_pct = category_epsilon_pct
+
         self.verbose = verbose
         self.loss = loss
 
@@ -167,10 +171,25 @@ class PATECTGAN(CTGANSynthesizer):
 
         self._device = torch.device(device)
 
+        if self._log_frequency:
+            warnings.warn(
+                "log_frequency is selected.  This may result in oversampling frequent "
+                "categories, which could cause privacy leaks."
+            )
+
     def train(self, data, categorical_columns=None, ordinal_columns=None, update_epsilon=None):
         if update_epsilon:
             self.epsilon = update_epsilon
 
+        for col in categorical_columns:
+            if str(data[col].dtype).startswith('float'):
+                raise ValueError(
+                    "It looks like you are passing in a vector of continuous values"
+                    f"to a categorical column at [{col}]."
+                    "Please discretize and pass in categorical columns with"
+                    "unsigned integer or string category names."
+                )
+
         sample_per_teacher = (
             self.sample_per_teacher if self.sample_per_teacher < len(data) else 1000
         )
@@ -185,18 +204,45 @@ class PATECTGAN(CTGANSynthesizer):
 
         data_dim = self._transformer.output_dimensions
 
+        sampler_eps = 0.0
+        if categorical_columns and self._category_epsilon_pct:
+            sampler_eps = self.epsilon * self._category_epsilon_pct
+            per_col_sampler_eps = sampler_eps / len(categorical_columns)
+            self.epsilon = self.epsilon - sampler_eps
+        else:
+            per_col_sampler_eps = None
+
         self.cond_generator = DataSampler(
             train_data,
             self._transformer.output_info_list,
-            self._log_frequency)
+            self._log_frequency,
+            per_column_epsilon=per_col_sampler_eps
+        )
 
+        spent = self.cond_generator.total_spent
+        if (
+            spent > sampler_eps
+            and not np.isclose(spent, sampler_eps)
+        ):
+            raise AssertionError(
+                f"The data sampler used {spent} epsilon and was budgeted for {sampler_eps}"
+            )
         # create conditional generator for each teacher model
 
         # Note: Previously, there existed a ConditionalGenerator object in CTGAN
         # - that functionality has been subsumed by DataSampler, but switch is
         # essentially 1 for 1
+        # don't need to count eps for each teacher, because these are disjoint partitions
+        cached_probs = self.cond_generator.discrete_column_category_prob
+
         cond_generator = [
-            DataSampler(d, self._transformer.output_info_list, self._log_frequency)
+            DataSampler(
+                d,
+                self._transformer.output_info_list,
+                self._log_frequency,
+                per_column_epsilon=None,
+                discrete_column_category_prob=cached_probs
+            )
             for d in data_partitions
         ]
 
@@ -239,7 +285,7 @@ class PATECTGAN(CTGANSynthesizer):
         noise_multiplier = self.noise_multiplier
         alphas = torch.tensor([0.0 for i in range(self.moments_order)], device=self._device)
         l_list = 1 + torch.tensor(range(self.moments_order), device=self._device)
-        eps = 0
+        eps = torch.zeros(1)
 
         mean = torch.zeros(self._batch_size, self._embedding_dim, device=self._device)
         std = mean + 1
@@ -252,12 +298,35 @@ class PATECTGAN(CTGANSynthesizer):
         if self.verbose:
             print("using loss {} and regularization {}".format(self.loss, self.regularization))
 
-        while eps < self.epsilon:
+        iteration = 0
+
+        if self.delta is None:
+            self.delta = 1 / (train_data.shape[0] * np.sqrt(train_data.shape[0]))
+
+        while eps.item() < self.epsilon:
+            iteration += 1
+
+            eps = min((alphas - math.log(self.delta)) / l_list)
+
+            if eps.item() > self.epsilon:
+                if iteration == 1:
+                    raise ValueError(
+                                "Inputted epsilon parameter is too small to"
+                                + " create a private dataset. Try increasing epsilon and rerunning."
+                            )
+                break
+
             # train teacher discriminators
             for t_2 in range(self.teacher_iters):
                 for i in range(self.num_teachers):
                     partition_data = data_partitions[i]
-                    data_sampler = DataSampler(partition_data, self._transformer.output_info_list, self._log_frequency)
+                    data_sampler = DataSampler(
+                        partition_data,
+                        self._transformer.output_info_list,
+                        self._log_frequency,
+                        per_column_epsilon=None,
+                        discrete_column_category_prob=cached_probs
+                    )
                     fakez = torch.normal(mean, std=std).to(self._device)
 
                     condvec = cond_generator[i].sample_condvec(self._batch_size)
@@ -315,7 +384,13 @@ class PATECTGAN(CTGANSynthesizer):
             ###
             # train student discriminator
             for t_3 in range(self.student_iters):
-                data_sampler = DataSampler(train_data, self._transformer.output_info_list, self._log_frequency)
+                data_sampler = DataSampler(
+                    train_data,
+                    self._transformer.output_info_list,
+                    self._log_frequency,
+                    per_column_epsilon=None,
+                    discrete_column_category_prob=cached_probs
+                )
                 fakez = torch.normal(mean=mean, std=std)
 
                 condvec = self.cond_generator.sample_condvec(self._batch_size)
@@ -417,8 +492,6 @@ class PATECTGAN(CTGANSynthesizer):
             loss_g.backward()
             optimizerG.step()
 
-            eps = min((alphas - math.log(self.delta)) / l_list)
-
             if self.verbose:
                 print(
                     "eps: {:f} \t G: {:f} \t D: {:f}".format(
